/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Quadratic Monte Carlo Method                                           //
//                                                                         //
//  B. Militzer                                     Berkeley 02-15-22      //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#include "Random.h"
#include "ParseCommandLineArguments.h"
#include "MC.h"
#include "Timer.h"
#include "Analysis.h"
#include "Blocker.h"
#include "Autocorrelation.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef HARMONIC

class QMCState {
 public:
  int n;
  Array1 <double> p;  
  Array1 <double> dP; // step sizes
  double y;

  QMCState():n(0){
    Reset();
  }
  QMCState(const int n_){
    Resize(n_);
    Reset();
  }
  void Resize(const int nn) {
    n = nn;
    p.Resize(nn);
    dP.Resize(nn);
  }

  int nDim() const {
    return n;
  }
  bool Valid(const bool printFlag=true) const {
    return true;
  }
  void Reset() {
    y = Evaluate();
  }
  double operator[](const int i) const {
    return p[i];
  }
  double & operator[](const int i) {
    return p[i];
  }
  double Parameter(const int i) const {
    return p[i];
  }
  double & Parameter(const int i) {
    return p[i];
  }
  double ParameterStep(const int i) const {
    return dP[i];
  }
  double & ParameterStep(const int i) {
    return dP[i];
  }

  double Evaluate() {
    return (*this)();
  }
  double operator()() {
    double s2 = 0.0;
    for(int i=0; i<n; i++) {
      s2 += sqr(p[i]);
      //      s2 += pow(abs(p[i]),2.5);
    }
    y = s2;
    return s2;
  }

  void PrintParametersAndStepSizes() const {
    for(int i=0; i<nDim(); i++) {
      cout << " QMC state parameter " << IntToStringMaxNumber(i+1,nDim()) << " = " << Parameter(i) << "    step= " << ParameterStep(i) << endl;
    }
  }


  friend ostream & operator<<(ostream & os, const QMCState & s) {
    os << s.p;
    return os;
  }

};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#else // HARMONIC

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class QMCState { // ring potential
 public:
  int n;
  Array1 <double> p;  
  Array1 <double> dP; // step sizes
  double y;
  double R, m, m2, B;

  QMCState():n(0) {
    InitializeConstants();
    Reset();
  }
  QMCState(const int n_) {
    Resize(n_);
    InitializeConstants();
    Reset();
  }
  QMCState(const int n_, const double R_, const double m_, const double B_):R(R_),m(m_),m2(2.0*m),B(B_) {
    Resize(n_);
    Reset();
  }
  void InitializeConstants() {
    R = 1.0;
    m = 1.0;
    m2= 2.0*m;
    B = 0.01;
  }
  void Resize(const int n_) {
    n = n_;
    p.Resize(n_);
    dP.Resize(n_);
    //    Write3(n,p,dP);
  }

  int nDim() const {
    return n;
  }
  bool Valid(const bool printFlag=true) const {
    return true;
  }
  void Reset() {
    Evaluate();
  }
  double operator[](const int i) const {
    return p[i];
  }
  double & operator[](const int i) {
    return p[i];
  }
  double Parameter(const int i) const {
    return p[i];
  }
  double & Parameter(const int i) {
    return p[i];
  }
  double ParameterStep(const int i) const {
    return dP[i];
  }
  double & ParameterStep(const int i) {
    return dP[i];
  }

  double operator()() { // This is where we calculate the potential, return value and store it in 'y'
    if (n<2) return 0.0; // n=0 can happen during default constructor called with Array1<QMCState> a.Resize(10);
    double rho = sqrt(sqr(p[0])+sqr(p[1])); // rho^2 = x^2 + y^2
    //    y = pow(rho-R,m2);
    //    for(int i=2; i<n; i++) { // all parameters except x and y
    //      y += pow(p[i],m2);
    //    }
    y = pow( m2*(rho-R) , m2);
    for(int i=2; i<n; i++) { // all parameters except x and y
      y += pow( m2*p[i] , m2);
    }

    y -= B*p[0]; 
    return y;
  }
  double Evaluate() {
    return (*this)();
  }

  void PrintParametersAndStepSizes() const {
    for(int i=0; i<nDim(); i++) {
      cout << " QMC state parameter " << IntToStringMaxNumber(i+1,nDim()) << " = " << Parameter(i) << "    step= " << ParameterStep(i) << endl;
    }
  }


  friend ostream & operator<<(ostream & os, const QMCState & s) {
    os << "R=" << s.R << " m= " << s.m << " B= " << s.B << " y= " << s.y << " r= " << s.p;
    return os;
  }

};

#endif // HARMONIC

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

template <class S> 
int SetUpSimplex(S & s0, Array1 <S> & s, const double minDeltaYAtStart, const double minDeltaYFractionAtStart, const double maxDeltaYAtStart, const double maxDeltaYFractionAtStart) {
  const bool print = false;
  int nDim     = s0.nDim();
  if (print) cout << "Setting up simplex in " << nDim << " dimensions from single state." << endl;

  if (!s0.Valid()) {
    cout << s0 << endl;
    s0.Valid(true);
    error("Cannot set up simplex starting with an invalid parameter set.");
    //    warning("Should not start MC with an invalid parameter set. May be a benign S2<S1 issue.");
  }

  int nWalkers = nDim+1;
  s.Resize(nWalkers); 
  
  s[0] = s0;       // State class needs copy properly. 
  s[0].Evaluate(); // evaluate s0, this takes time
  int nFunc = 1;
  double y0 = s[0].y;
  if (print) Write(y0);
  if (print) s[0].PrintParametersAndStepSizes();

  bool minDYAtStartFlag = (minDeltaYAtStart>0.0) || (minDeltaYFractionAtStart>0.0); // true if we want to try to generate an ensemble with similar y values
  double minDYAtStart1 = minDeltaYFractionAtStart*abs(s[0].y); // y0 may be less than 0 but it always provides a scale. 
  double minDYAtStart2 = minDeltaYAtStart;                     
  double minDYAtStart  = max(minDYAtStart1,minDYAtStart2); // take the larger of the two because do the same below

  bool maxDYAtStartFlag = (maxDeltaYAtStart>0.0) || (maxDeltaYFractionAtStart>0.0); // true if we want to try to generate an ensemble with similar y values
  double maxDYAtStart1 = maxDeltaYFractionAtStart*abs(s[0].y); // y0 may be less than 0 but it always provides a scale. 
  double maxDYAtStart2 = maxDeltaYAtStart;                     
  double maxDYAtStart  = max(maxDYAtStart1,maxDYAtStart2); // take the larger of the two. This solve the issue of y~1e-5 and if maxDeltaYFractionAtStart or maxDeltaYAtStart are negative. If both are <0, we need to use maxYAtStartFlag 

  for(int i=0; i<nDim; i++) { // set up initial tetrahedron of nDim points in addition to original one
    s[i+1] = s0; // here we make copies. Make sure the state copies properly.

    double step         = s0.ParameterStep(i);
    double previousStep = step;
    double previousY    = y0;
    int k=0;
    bool reduceDY = true;
    while (true) {
      k++; // increases counter k
      //  pp(i+1,i) = s[i] + step; // do not use s[i] because it may have changed by the call 's(pp[i+1])'
      s[i+1].Parameter(i) = s0.Parameter(i) + step;
      s[i+1].Evaluate(); 
      nFunc++;
      // y[i+1]    = s(pp[i+1]); // warning, depending how State is written, this call may copy pp[i+1] over into state 's' !!!!!!!!!!!
      double dy = s[i+1].y - y0;
      if (print) { cout << " ::::::"; Write8(i+1,k,step,y0,s[i+1].y,dy,minDYAtStart,maxDYAtStart); }
      //      cout << endl; // print s[i+1] last because the line is sooo long

      if (dy<0.0) { // Surprise, we found a new minimum. Keep it and make it the new state to compare with. (May be lost otherwise.)
	s[0]   = s[i+1];   // needs to be copied into s[0], not into s0, to be used in simplex method
	s[0].y = s[i+1].y;
	y0     = s[i+1].y; // since we write 'y0' for convenience, overwrite this value also
	step *=  +2.0;
	k--; // reduce so that we do one more step after this one. Do not reset k to 0 because the this routine may never terminate otherwise. 
	continue;
      }

      bool dYTooSmall = minDYAtStartFlag && (abs(dy)<minDYAtStart);
      bool dYTooLarge = maxDYAtStartFlag && (dy     >maxDYAtStart); // large negative dY are ok
      if (k==1) reduceDY = dYTooLarge; // decide once (in the first step) whether we need to increase or decrease the step size

      if (!dYTooSmall && !dYTooLarge)      break; // success, not too large, not too small
      if (k>1 &&  reduceDY && !dYTooLarge) break; // success, was too large before but not anymore (do not check if it became too small now)
      if (k>1 && !reduceDY && !dYTooSmall) {      // success, was too small before but not anymore
	if (dYTooLarge) { // if it became too large, go back one step (requires k>1)
	  s[i+1].Parameter(i) = s0.Parameter(i) + previousStep;
	  s[i+1].y            = previousY;
	}
	break; 
      }

      const int kMax = 5;
      if (k>=kMax) {
	cout << " Could not adjust step size that dY became of the right magnitude. Breaking out anyway." << endl;
	if (reduceDY==false && dYTooLarge) { // if we tried increase dY but the last step went too far, go back to previous step
	  s[i+1].Parameter(i) = s0.Parameter(i) + previousStep;
	  s[i+1].y            = previousY;
	}
	break;
      }

      previousY    = s[i+1].y;
      previousStep = step;
      if (reduceDY) step *=  -0.1; // change direction and reduce step size quite a bit and try again	
      else          step *=  +2.0;
    }
  }
  return nFunc;
}

// see http://extremelearning.com.au/how-to-generate-uniformly-random-points-on-n-spheres-and-n-balls/
void SampleNSphere(Array1 <double> & x, const int n) {
  double d = 0.0;
  for(int i=0; i<n; i++) {
    x[i] = SampleGaussian(1.0); // Box-Mueller formula in MC.h
    d += sqr(x[i]);
  }
  d = 1.0/sqrt(d);
  for(int i=0; i<n; i++) {
    x[i] *= d;
  }
}

template <class S> 
int SetUpNWalkersGeneral(S & s0, Array1 <S> & s, const int nWalkers, const double minDeltaYAtStart, const double minDeltaYFractionAtStart, const double maxDeltaYAtStart, const double maxDeltaYFractionAtStart) {
  //  const bool print = false;
  const bool print = true;
  int nDim     = s0.nDim();
  if (print) cout << "Setting up " << nWalkers << " walkers in " << nDim << " dimensions from single state." << endl;

  if (!s0.Valid()) {
    cout << s0 << endl;
    s0.Valid(true);
    error("Cannot set up simplex starting with an invalid parameter set.");
    //    warning("Should not start MC with an invalid parameter set. May be a benign S2<S1 issue.");
  }

  //  int nWalkers = nDim+1;
  s.Resize(nWalkers); 
  
  s[0] = s0;       // State class needs copy properly. 
  s[0].Evaluate(); // evaluate s0, this takes time
  int nFunc = 1;
  double y0 = s[0].y;
  if (print) Write(y0);
  if (print) s[0].PrintParametersAndStepSizes();

  bool minDYAtStartFlag = (minDeltaYAtStart>0.0) || (minDeltaYFractionAtStart>0.0); // true if we want to try to generate an ensemble with similar y values
  double minDYAtStart1 = minDeltaYFractionAtStart*abs(s[0].y); // y0 may be less than 0 but it always provides a scale. 
  double minDYAtStart2 = minDeltaYAtStart;                     
  double minDYAtStart  = max(minDYAtStart1,minDYAtStart2); // take the larger of the two because do the same below

  bool maxDYAtStartFlag = (maxDeltaYAtStart>0.0) || (maxDeltaYFractionAtStart>0.0); // true if we want to try to generate an ensemble with similar y values
  double maxDYAtStart1 = maxDeltaYFractionAtStart*abs(s[0].y); // y0 may be less than 0 but it always provides a scale. 
  double maxDYAtStart2 = maxDeltaYAtStart;                     
  double maxDYAtStart  = max(maxDYAtStart1,maxDYAtStart2); // take the larger of the two. This solve the issue of y~1e-5 and if maxDeltaYFractionAtStart or maxDeltaYAtStart are negative. If both are <0, we need to use maxYAtStartFlag 

  Array1 <double> sphere(nDim);
  for(int i=0; i<nWalkers-1; i++) { // set up initial tetrahedron of nDim points in addition to original one
    s[i+1] = s0; // here we make copies. Make sure the state copies properly.

    SampleNSphere(sphere,nDim);
    for(int j=0; j<nDim; j++) { 
      sphere[j] *=  s0.ParameterStep(j); // scale every direction
    }

    double step         = 1.0;
    double previousStep = step;
    double previousY    = y0;
    int k=0;
    bool reduceDY = true;
    while (true) {
      k++; // increases counter k
      //  pp(i+1,i) = s[i] + step; // do not use s[i] because it may have changed by the call 's(pp[i+1])'
      for(int j=0; j<nDim; j++) { 
	s[i+1].Parameter(j) = s0.Parameter(j) + sphere[j]*step;
      }
      s[i+1].Evaluate(); 
      nFunc++;
      // y[i+1]    = s(pp[i+1]); // warning, depending how State is written, this call may copy pp[i+1] over into state 's' !!!!!!!!!!!
      double dy = s[i+1].y - y0;
      if (print) { cout << " ::::::"; Write8(i+1,k,step,y0,s[i+1].y,dy,minDYAtStart,maxDYAtStart); }
      //      cout << endl; // print s[i+1] last because the line is sooo long

      if (dy<0.0) { // Surprise, we found a new minimum. Keep it and make it the new state to compare with. (May be lost otherwise.)
	s[0]   = s[i+1];   // needs to be copied into s[0], not into s0, to be used in simplex method
	s[0].y = s[i+1].y;
	y0     = s[i+1].y; // since we write 'y0' for convenience, overwrite this value also
	step *=  +2.0;
	k--; // reduce so that we do one more step after this one. Do not reset k to 0 because the this routine may never terminate otherwise. 
	continue;
      }

      bool dYTooSmall = minDYAtStartFlag && (abs(dy)<minDYAtStart);
      bool dYTooLarge = maxDYAtStartFlag && (dy     >maxDYAtStart); // large negative dY are ok
      if (k==1) reduceDY = dYTooLarge; // decide once (in the first step) whether we need to increase or decrease the step size

      if (!dYTooSmall && !dYTooLarge)      break; // success, not too large, not too small
      if (k>1 &&  reduceDY && !dYTooLarge) break; // success, was too large before but not anymore (do not check if it became too small now)
      if (k>1 && !reduceDY && !dYTooSmall) {      // success, was too small before but not anymore
	if (dYTooLarge) { // if it became too large, go back one step (requires k>1)
	  for(int j=0; j<nDim; j++) { 
	    s[i+1].Parameter(j) = s0.Parameter(j) + sphere[j]*previousStep;
	  }
	  s[i+1].y            = previousY;
	}
	break; 
      }

      const int kMax = 5;
      if (k>=kMax) {
	cout << " Could not adjust step size that dY became of the right magnitude. Breaking out anyway." << endl;
	if (reduceDY==false && dYTooLarge) { // if we tried increase dY but the last step went too far, go back to previous step
	  for(int j=0; j<nDim; j++) { 
	    s[i+1].Parameter(j) = s0.Parameter(j) + sphere[j]*previousStep;
	  }
	  s[i+1].y            = previousY;
	}
	break;
      }

      previousY    = s[i+1].y;
      previousStep = step;
      if (reduceDY) step *=  -0.1; // change direction and reduce step size quite a bit and try again	
      else          step *=  +2.0;
    }
  }
  return nFunc;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <class S> 
void RunMC(S & s0, Array1 <S> s, const double a, const int nWalkers, const double temp, const int nBlocks, const bool flagWalk, const bool flagAffine, const bool flagQMC, const bool flagGaussian, Array1 <ScalarAverage> & energy) {
  //  const bool print = false;
  const bool print = true;
  const int blockMin = nBlocks / 20;
  Analysis analysis(blockMin);

  //  int nWalkers = s0.nDim() + 1;
  //  if (nWalkers != s.Size()) error("RunQuadraticMC: nWalkers != s.Size()");
  if (nWalkers<3) error("nWalkers too small",nWalkers);

  const string line = " -----------------------------------------------------------------------------------";
  if (print) cout << endl << " ======= Starting Monte Carlo run with " << s0.nDim() << " dimensions and " << nWalkers << " walkers. =======" << endl << endl;

  for(int j=0; j<s.Size(); j++) {
    analysis.AddTotalEnergy(s[j].y);
    energy[0].Add(s[j].y);
  }

  //  double a = 2.5;
  //  double a = 5.0;
  //  double a = 3.0;
  //  double a = 1.0; // good for QMC
  //  double a = 0.9; // best for QMC
  //  double a = 0.95;
  for(int iBlock=0; iBlock<nBlocks; iBlock++) { 
    //    if (print) cout << line << endl; 

    analysis.StartBlock(iBlock);
    AccRatio accRatio;
    //    const int nSteps = 10; // steps per block
    //    const int nSteps = 100; // steps per block
    //    const int nSteps = 1000; // steps per block
    const int nSteps = 1000.0 / double(nWalkers); // steps per block
    //    const int nSteps = 100.0 / double(nWalkers); // steps per block
    for (int iStep=0; iStep<nSteps; iStep++) {

      //      if (Random()<0.01) PerformOneQuadraticMonteCarloSweep(s,accRatio,a,nWalkers,temp); 
      if (flagWalk)   PerformWalkMoves(s,accRatio,nWalkers,temp); 
      if (flagAffine) PerformOneAffineMonteCarloSweep(s,accRatio,a,nWalkers,temp); 
      if (flagQMC)    PerformOneQuadraticMonteCarloSweep(s,accRatio,a,nWalkers,temp,flagGaussian); 

      for(int j=0; j<s.Size(); j++) {
	analysis.AddTotalEnergy(s[j].y);
	energy[iBlock+1].Add(s[j].y);
      }
    }
    if (print) {
      cout << " " << IntToStringMaxNumber(iBlock,nBlocks) << " : ";
      //      cout << accRatio;
      cout << "ratio= " << fix(6,4,accRatio.Ratio());

      //      cout << IntToStringMaxNumber(iBlock,nBlocks) << " : ";
      //      analysis.EndBlock(iBlock,print);
      analysis.EndBlock(iBlock,false);
      analysis.PrintBlockEnergy();
      cout << endl;

      //    Quit("after first MC sweep");
    }
    if (a<1e-8) error("Scaling parameters 'a' became too small",a); // added on 5/11/20 to prevent a from reaching 1e-20

  }

  if (print) {
    //  if (true) {
    cout << line << endl; 
    cout << line << endl; 
    analysis.EndSimulation();
    cout << endl;
    cout << " Last step:" << endl;
    cout << endl;
    //  cout << s;
    for(int i=0; i<s.Size(); i++) {
      cout << " state " <<  IntToStringMaxNumber(i,s.Size()) << " : " << s[i] << endl;
    }
    cout << endl << " ======= End of MC chain ======= " << endl << endl;
  }
}

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

inline Array1 <double> ExtractArray(const Array1 <ScalarAverage> & sa) {
  const int n = sa.Size();
  Array1 <double> x(n);
  for(int i=0; i<n; i++) {
    x[i] = sa[i].GetAverage();
  }
  return x;
}

void BlockingAnalysis(const Array1 <ScalarAverage> & sa) {
  cout << endl;

  Array1 <double> x = ExtractArray(sa);
  const int n2 = x.Size();
  
  int nF = 10;
  for(int iF=nF; iF>0; iF--) { // iF==1 means using the last 10%
    int n1 = n2 - (n2*iF) / nF; // integer division
    int n  = n2 - n1;
    //    Write4(iF,nF,n1,n);
    double av,err;
    Form f(fix,14,8);
    //    double err1; Blocker(x,n1,n2,av,err,err1,true); // debug statement to compare with 'blocker04.c'
    BlockerFast(x,n1,n2,av,err);
    cout << " Fraction= " << fix(3,1,double(iF)/double(nF)) 
	 << " n= " << IntToStringMaxNumber(n,n2)
	 << " E= " << f(av) << " +- " << f(err) 
	 << endl;
  }
}

void AutocorrelationAnalysis(const Array1 <ScalarAverage> & sa) {
  cout << endl;

  Array1 <double> x = ExtractArray(sa);
  const int n2 = x.Size();
  
  const int nF = 10;
  for(int iF=nF; iF>0; iF--) {        // study different fraction starting from end
    const int n1 = n2 - (n2*iF) / nF; // integer division
    const int nn = n2 - n1;           // number of data points to be considered here

    const int nCorrMax = nn / 10; // integer division
    double sum = 0.0;
    double c0 = Autocorrelation(x,0,n1,n2);
    int step = 1;
    int k = 0;
    for(int iCorr=0; iCorr<nCorrMax; iCorr+=step) {
      double c = Autocorrelation(x,iCorr,n1,n2) / c0;
      sum += c;
      cout << " f= " << fix(3,1,double(iF)/double(nF)) 
	   << " step= " << IntToStringMaxNumber(iCorr,nCorrMax) 
	   << " ac= " << c 
	   << " sum= " << sum << endl;
      if ((++k)%10==0) step *= 2; // double step size every 10 steps
    }
  }
}

/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv) {
  cout << endl;
  cout << "Starting a QMC calculation." << endl;
  cout << endl;
  cout << "Process ID: " << getpid() << endl;
  cout << endl;

  Array1 <string> arg = CopyCommandLineArgumentsToArray(argc,argv);
  cout.precision(10);

  if (ParseCommandLineFlagWithOrWithoutDash(arg,"noRandom")) {
    //  InitRandom(false); warning("Using standard random numbers");
    InitRandom(false);
    cout << "Using standard random numbers stream: ";
  } else {
    InitRandom(true);
    cout << "Using randomized random numbers stream: ";
  }
  Write(Random());

  //  int n = 10;
  int n = 3;
  ParseCommandLineArgumentNoSpace(arg,"n=",n);
  Write(n);

  double R = 1.0;
  ParseCommandLineArgumentNoSpace(arg,"R=",R);
  double m = 8.0;
  ParseCommandLineArgumentNoSpace(arg,"m=",m);
  double B = 0.01;
  ParseCommandLineArgumentNoSpace(arg,"B=",B);
  Write3(R,m,B);

  //  double temp = 0.1;
  //  double temp = 0.01; // so that the default of maxDeltaYAtStart becomes 0.01
  //  double temp = 1.0;
  double temp = 0.001;
  ParseCommandLineArgumentNoSpace(arg,"T=",temp);
  double a = 1.5;
  ParseCommandLineArgumentNoSpace(arg,"a=",a);
  //  if (a<=1.0) error("'a' too small",a);
  Write2(temp,a);

  double maxDeltaYFractionAtStart = temp;     
  double minDeltaYFractionAtStart = maxDeltaYFractionAtStart/10.0; 

  double maxDeltaYAtStart         = temp;      
  double minDeltaYAtStart         = temp/10.0; 

  bool flagWalk = false;
  flagWalk = ParseCommandLineFlag(arg,"-walk");
  bool flagAffine = false;
  flagAffine = ParseCommandLineFlag(arg,"-affine");
  bool flagQMC = false;
  flagQMC = ParseCommandLineFlag(arg,"-qmc");
  bool flagGaussian = false;
  flagGaussian = ParseCommandLineFlag(arg,"-Gaussian");

  ParseCommandLineArgumentNoSpace(arg,"maxDeltaYFractionAtStart=",maxDeltaYFractionAtStart);
  ParseCommandLineArgumentNoSpace(arg,"maxDeltaYAtStart=",maxDeltaYAtStart);
  Write2(maxDeltaYFractionAtStart,maxDeltaYAtStart);
  ParseCommandLineArgumentNoSpace(arg,"minDeltaYFractionAtStart=",minDeltaYFractionAtStart);
  ParseCommandLineArgumentNoSpace(arg,"minDeltaYAtStart=",minDeltaYAtStart);
  Write2(minDeltaYFractionAtStart,minDeltaYAtStart);

  int nRuns   = 1;
  //  int nRuns   = 100;
  ParseCommandLineArgumentNoSpace(arg,"nRuns=",nRuns);

  //  int nBlocks = 250;
  int nBlocks = 1000;
  ParseCommandLineArgumentNoSpace(arg,"nBlocks=",nBlocks);

#ifndef HARMONIC
  bool lowFlag = false; // start from high-energy point by default
  lowFlag = ParseCommandLineFlag(arg,"-low");
#endif
  
  int nWalkers = -1;
  bool nWalkersFlag = ParseCommandLineArgumentNoSpace(arg,"nWalkers=",nWalkers);

  Array1 <ScalarAverage> energy(nBlocks+1);
  for(int iRun=0; iRun<nRuns; iRun++) {

#ifdef HARMONIC

    QMCState s0(n);
    for(int i=0; i<n; i++) {
      s0.Parameter(i)     = Random();
      s0.ParameterStep(i) = 0.1;
    }

#else // HARMONIC

    QMCState s0(n,R,m,B);
    for(int i=0; i<n; i++) {
      if (lowFlag==false) {
	s0.Parameter(i)     = (i>0) ? 0.0 : -R; // start from high-energy point
      } else {
	s0.Parameter(i)     = (i>0) ? 0.0 : +R; // start from low-energy point
      }
      s0.ParameterStep(i) = 0.1;
    }

#endif // HARMONIC

    Array1 <QMCState> s;
    if (nWalkersFlag==false) {
      SetUpSimplex(s0,s,minDeltaYAtStart,minDeltaYFractionAtStart,maxDeltaYAtStart,maxDeltaYFractionAtStart);
      nWalkers = s.Size();
    } else {
      SetUpNWalkersGeneral(s0,s,nWalkers,minDeltaYAtStart,minDeltaYFractionAtStart,maxDeltaYAtStart,maxDeltaYFractionAtStart);
    }

    RunMC(s0,s,a,nWalkers,temp,nBlocks,flagWalk,flagAffine,flagQMC,flagGaussian,energy);

    AutocorrelationAnalysis(energy);
    BlockingAnalysis(energy);
  }

  //  cout << endl;

  /*
  for(int iBlock=0; iBlock<=nBlocks; iBlock++) {
    Form f(fix,12,8);
    cout << " Block= " << IntToStringMaxNumber(iBlock,nBlocks) << " E= " << f(energy[iBlock].GetAverage()) << " var= " << f(energy[iBlock].GetVariance()) << endl;
  }
  cout << endl;
  Quit("Finished simple MC run.");
  */

  cout << endl << " Program finished properly. Happy QMC computing!" << endl;
  return 0;
}
